import argparse
import json
import re
import traceback
import typing

import cf_xarray  # noqa: F401
import pydantic
import pydantic_core
import requests
import upath
import xarray as xr
from ruamel.yaml import YAML

yaml = YAML(typ='safe')


def s3_to_https(s3_url: str) -> str:
    # Split the URL into its components
    s3_parts = s3_url.split('/')

    # Get the bucket name from the first part of the URL
    bucket_name = s3_parts[2]

    # Join the remaining parts of the URL to form the path to the file
    path = '/'.join(s3_parts[3:])

    # Return the HTTPS URL in the desired format
    return f'https://{bucket_name}.s3.amazonaws.com/{path}'


def gs_to_https(gs_url: str) -> str:
    return gs_url.replace('gs://', 'https://storage.googleapis.com/')


class XarrayOpenKwargs(pydantic.BaseModel):
    engine: typing.Literal['zarr', 'kerchunk']


default_xarray_open_kwargs = XarrayOpenKwargs(engine='zarr')


class RechunkingItem(pydantic.BaseModel):
    path: str = pydantic.Field(..., description='Path to the rechunked store')
    use_case: str = pydantic.Field(..., description='Use case of the rechunking')


class Store(pydantic.BaseModel):
    id: str = pydantic.Field(..., description='ID of the store')
    name: str = pydantic.Field(None, description='Name of the store')
    url: str = pydantic.Field(..., description='URL of the store')
    rechunking: list[RechunkingItem] | None = pydantic.Field(None)
    public: bool | None = pydantic.Field(None, description='Whether the store is public')
    geospatial: bool | None = pydantic.Field(None, description='Whether the store is geospatial')
    xarray_open_kwargs: XarrayOpenKwargs | None = pydantic.Field(
        default_xarray_open_kwargs, description='Xarray open kwargs for the store'
    )
    last_updated: str | None = pydantic.Field(None, description='Last updated timestamp')


class Link(pydantic.BaseModel):
    label: str = pydantic.Field(..., description='Label of the link')
    url: str = pydantic.Field(..., description='URL of the link')


class LicenseLink(pydantic.BaseModel):
    title: str = pydantic.Field(..., description='Name of the license')
    url: str | None = pydantic.Field(None, description='URL of the license')


class Maintainer(pydantic.BaseModel):
    name: str = pydantic.Field(..., description='Name of the maintainer')
    github: str | None = pydantic.Field(None, description='GitHub username of the maintainer')


class Provider(pydantic.BaseModel):
    name: str = pydantic.Field(..., description='Name of the provider')
    description: str | None = pydantic.Field(None, description='Description of the provider')
    roles: list[str] | None = pydantic.Field(None, description='Roles of the provider')
    url: str | None = pydantic.Field(None, description='URL of the provider')


class Provenance(pydantic.BaseModel):
    providers: list[Provider]
    license: str
    license_link: LicenseLink | None = None


class Feedstock(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(validate_assignment=True)

    title: str = pydantic.Field(..., description='Title of the feedstock')
    description: str = pydantic.Field(..., description='Description of the feedstock')
    maintainers: list[Maintainer]
    provenance: Provenance
    thumbnail: pydantic.HttpUrl | None = pydantic.Field(
        None, description='Thumbnail of the feedstock'
    )
    tags: list[str] | None = pydantic.Field(None, description='Tags of the dataset')
    links: list[Link] | None = None
    stores: list[Store] | None = None
    meta_yaml_url: pydantic.HttpUrl | None = pydantic.Field(
        None, alias='ncviewjs:meta_yaml_url', description='URL of the meta YAML'
    )

    @classmethod
    def from_yaml(cls, path: str):
        content = yaml.load(upath.UPath(path).read_text())

        meta_url_key = next(
            (key for key in ['meta_yaml_url', 'ncviewjs:meta_yaml_url'] if key in content), None
        )

        if meta_url_key:
            meta_url = convert_to_raw_github_url(content[meta_url_key])
            meta = yaml.load(upath.UPath(meta_url).read_text())
            content = content | meta
        return cls.model_validate(content)


def convert_to_raw_github_url(github_url):
    # Check if the URL is already a raw URL
    if 'raw.githubusercontent.com' in github_url:
        return github_url

    # Replace the domain
    raw_url = github_url.replace('github.com', 'raw.githubusercontent.com')

    # Remove '/blob'
    raw_url = raw_url.replace('/blob', '')

    return raw_url


class ValidationError(Exception):
    def __init__(self, errors: list[dict[str, str]] | str) -> None:
        self.errors = errors
        super().__init__(self.errors)


def collect_feedstocks(path: upath.UPath) -> list[str]:
    """Collects all the datasets in the given directory."""

    url = convert_to_raw_github_url(path)
    if not (feedstocks := yaml.load(upath.UPath(url).read_text())['feedstocks']):
        raise FileNotFoundError(f'No YAML files (.yaml or .yml) found in {path}')
    return feedstocks


def format_report(title: str, feedstocks: list[dict], include_traceback: bool = False) -> str:
    report = f'{title} ({len(feedstocks)})\n'
    if not feedstocks:
        report += '  🚀 None found\n'
    else:
        for entry in feedstocks:
            report += f"  📂 {entry['feedstock']}\n"
            if include_traceback:
                report += f"    🔎 {entry['traceback']}\n"
    return report


def get_http_url(store: str) -> str:
    if store.startswith('s3://'):
        url = s3_to_https(store)

    elif store.startswith('gs://'):
        url = gs_to_https(store)
    else:
        url = store

    url = url.strip('/')
    return url


def is_store_public(store: str) -> bool:
    try:
        url = get_http_url(store)
        path = f'{url}/.zmetadata'

        response = requests.get(path)
        response.raise_for_status()
        return True
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 404:
            print(f'Resource not found at {path}.')
        else:
            print(f'HTTP error {e.response.status_code} for {path}.')
        return False
    except Exception as e:
        print(f'An error occurred while checking if store {store} is public: {str(e)}')
        return False


def load_store(store: str, engine: str) -> xr.Dataset:
    url = get_http_url(store)
    return xr.open_dataset(url, engine=engine, chunks={}, decode_cf=False)


def is_geospatial(ds: xr.Dataset, is_multiscale: bool) -> bool:
    if is_multiscale:
        return 'x' in ds.dims and 'y' in ds.dims
    cf_axes = ds.cf.axes

    # Regex patterns that match 'lat', 'latitude', 'lon', 'longitude' and also allow prefixes
    lat_pattern = re.compile(r'.*(lat|latitude)$', re.IGNORECASE)
    lon_pattern = re.compile(r'.*(lon|longitude)$', re.IGNORECASE)

    # Gather all coordinate and dimension names
    all_names = set(ds.coords.keys()).union(set(ds.dims))

    # Identify if both latitude and longitude coordinates/dimensions are present
    has_latitude = any(lat_pattern.match(name) for name in all_names)
    has_longitude = any(lon_pattern.match(name) for name in all_names)

    return ('X' in cf_axes and 'Y' in cf_axes) or (has_latitude and has_longitude)


def check_stores(feed: Feedstock) -> None:
    if feed.stores:
        for index, store in enumerate(feed.stores):
            print(f'  🚦 {store.id} ({index + 1}/{len(feed.stores)})')
            check_single_store(store)


def check_single_store(store: Store) -> None:
    multiscale_path = next(
        (entry.path for entry in store.rechunking or [] if entry.use_case == 'multiscales'),
        None,
    )
    is_public = is_store_public(multiscale_path or store.url)
    store.public = is_public
    if is_public:
        # check if the store is geospatial
        if multiscale_path:
            dt = xr.open_datatree(multiscale_path, engine='zarr', chunks={}, decode_cf=False)
            ds = dt['0'].ds
            is_geospatial_store = is_geospatial(ds, True)

        else:
            ds = load_store(
                store.url,
                store.xarray_open_kwargs.engine if store.xarray_open_kwargs else 'zarr',
            )
            is_geospatial_store = is_geospatial(ds, False)
        store.geospatial = is_geospatial_store
        # get last_updated_timestamp
        store.last_updated = ds.attrs.get('pangeo_forge_build_timestamp', None)


def validate_feedstocks(*, feedstocks: list[str]) -> list[Feedstock]:
    errors = []
    valid = []
    catalog = []

    for feedstock in feedstocks:
        try:
            feed = Feedstock.from_yaml(convert_to_raw_github_url(feedstock))
            if feed.stores:
                print('🔄 Checking stores')
                check_stores(feed)

            else:
                print('🚀 No stores found.')
            valid.append({'feedstock': str(feedstock), 'status': 'valid'})
            catalog.append(feed)
        except Exception:
            errors.append({'feedstock': str(feedstock), 'traceback': traceback.format_exc()})

    valid_report = format_report('✅ Valid feedstocks:', valid)
    invalid_report = format_report('❌ Invalid feedstocks:', errors, include_traceback=True)

    print(valid_report)
    print(invalid_report)
    print('\n\n')

    if errors:
        raise ValidationError('Validation failed')

    return catalog


def validate(args):
    if args.single:
        # If single file path is provided, validate just this one feedstock
        try:
            _ = Feedstock.from_yaml(convert_to_raw_github_url(args.single))
            print(
                format_report(
                    '✅ Valid feedstock:', [{'feedstock': str(args.single), 'status': 'valid'}]
                )
            )
        except Exception:
            print(
                format_report(
                    '❌ Invalid feedstock:',
                    [{'feedstock': str(args.single), 'traceback': traceback.format_exc()}],
                    include_traceback=True,
                )
            )
    else:
        # Default behavior, processing all feedstocks from directory
        feedstocks = collect_feedstocks(args.path)
        validate_feedstocks(feedstocks=feedstocks)


def generate(args):
    feedstocks = [args.single] if args.single else collect_feedstocks(args.path)
    catalog = validate_feedstocks(feedstocks=feedstocks)
    output = upath.UPath(args.output).resolve() / 'output'
    output.mkdir(parents=True, exist_ok=True)
    path = (
        output / 'single-feedstock-web-catalog.json'
        if args.single
        else output / 'consolidated-web-catalog.json'
    )
    with open(path, 'w') as f:
        json.dump(catalog, f, indent=2, default=pydantic_core.to_jsonable_python)
        print(f'Catalog written to {path}')


def main():
    parser = argparse.ArgumentParser(description='Utilities for cataloging feedstocks for LEAP')
    subparsers = parser.add_subparsers(help='sub-command help')

    # Subparser for the "validate" command
    parser_validate = subparsers.add_parser('validate', help='Validate the feedstocks')
    group = parser_validate.add_mutually_exclusive_group(required=True)
    group.add_argument('--path', type=str, help='Path to the feedstocks input YAML file')
    group.add_argument(
        '--single', type=str, help='Path to a single feedstock YAML file to validate'
    )
    parser_validate.set_defaults(func=validate)

    # Subparser for the "generate" command
    parser_generate = subparsers.add_parser('generate', help='Generate the catalog')
    group = parser_generate.add_mutually_exclusive_group(required=True)
    group.add_argument(
        '--path', type=str, help='Path to the feedstocks input YAML file or directory'
    )
    group.add_argument('--single', type=str, help='Path to a single feedstock YAML file')
    parser_generate.add_argument(
        '--output', type=str, required=True, help='Path to the output directory'
    )
    parser_generate.set_defaults(func=generate)

    args = parser.parse_args()
    if hasattr(args, 'func'):
        args.func(args)
    else:
        parser.print_help()


if __name__ == '__main__':
    main()



from ruamel.yaml import YAML

from leap_data_management_utils.data_management_transforms import (
    get_catalog_store_urls,
    get_github_commit_url,
)

yaml = YAML(typ='safe')


def test_smoke_test():
    assert True
    # This is a bit dumb, but it at least checks the the imports are working
    # again super hard to test code involving bigquery here.


def test_get_github_commit_url():
    url = get_github_commit_url()
    assert url.startswith('https://github.com/leap-stc/leap-data-management-utils')


def test_get_catalog_store_urls(tmp_path):
    # Create a temporary text file
    temp_file = tmp_path / 'some-name.yaml'
    data = {
        'stores': [{'id': 'a', 'url': 'a-url', 'some_other': 'stuff'}, {'id': 'b', 'url': 'b-url'}]
    }
    with open(temp_file, 'w') as f:
        yaml.dump(data, f)

    # Call the function to read the file
    content = get_catalog_store_urls(temp_file)

    # Assertions
    assert content['a'] == 'a-url'
    assert content['b'] == 'b-url'



# Note: All of this code was written by Julius Busecke and copied from this feedstock:
# https://github.com/leap-stc/cmip6-leap-feedstock/blob/main/feedstock/recipe.py#L262

import os
import subprocess
from dataclasses import dataclass
from datetime import datetime, timezone

import apache_beam as beam
import zarr
from ruamel.yaml import YAML

from leap_data_management_utils.bq_interfaces import BQInterface

yaml = YAML(typ='safe')


def get_github_actions_url() -> str:
    """Return the url of the gh action run"""
    if os.getenv('GITHUB_ACTIONS') == 'true':
        print('Running from within GH actions')
        server_url = os.getenv('GITHUB_SERVER_URL')
        repository = os.getenv('GITHUB_REPOSITORY')
        run_id = os.getenv('GITHUB_RUN_ID')
        commit_hash = os.getenv('GITHUB_SHA')

        if server_url and repository and run_id and commit_hash:
            return f'{server_url}/{repository}/actions/runs/{run_id}'
        else:
            print('One or more environment variables are missing.')
            return 'none'
    else:
        return 'none'


def get_github_commit_url() -> str:
    """Get the GitHub commit URL for the current commit"""
    # Get GitHub Server URL

    # check if this is running from within a github action
    if os.getenv('GITHUB_ACTIONS') == 'true':
        print('Running from within GH actions')
        server_url = os.getenv('GITHUB_SERVER_URL')
        repository = os.getenv('GITHUB_REPOSITORY')
        run_id = os.getenv('GITHUB_RUN_ID')
        commit_hash = os.getenv('GITHUB_SHA')

        if server_url and repository and run_id and commit_hash:
            git_url_hash = f'{server_url}/{repository}/commit/{commit_hash}'
        else:
            print(
                'Could not construct git_url_hash. One or more environment variables are missing.'
            )
            git_url_hash = 'none'

    else:
        # TODO: If the above fails, maybe still try this? Even though that would be a really rare case?
        print('Fallback: Calling git via subprocess')
        github_server_url = 'https://github.com'
        # Get the repository's remote origin URL
        try:
            repo_origin_url = subprocess.check_output(
                ['git', 'config', '--get', 'remote.origin.url'], text=True
            ).strip()

            # Extract the repository path from the remote URL
            repository_path = repo_origin_url.split('github.com/')[-1].replace('.git', '')

            # Get the current commit SHA
            commit_sha = subprocess.check_output(['git', 'rev-parse', 'HEAD'], text=True).strip()

            # Construct the GitHub commit URL
            git_url_hash = f'{github_server_url}/{repository_path}/commit/{commit_sha}'
        except Exception as e:
            print(f'Getting git_url_hash failed with {e}')
            git_url_hash = 'none'
    # Output the GitHub commit URL
    return git_url_hash


def get_catalog_store_urls(catalog_yaml_path: str) -> dict[str, str]:
    with open(catalog_yaml_path) as f:
        catalog_meta = yaml.load(f)
    return {d['id']: d['url'] for d in catalog_meta['stores']}


# ----------------------------------------------------------------------------------------------
# apache Beam stages
# ----------------------------------------------------------------------------------------------


@dataclass
class RegisterDatasetToCatalog(beam.PTransform):
    table_id: str
    dataset_id: str

    def _register_dataset_to_catalog(self, store: zarr.storage.FSStore) -> zarr.storage.FSStore:
        bq_interface = BQInterface(table_id=self.table_id)
        bq_interface.catalog_insert(dataset_id=self.dataset_id, dataset_url=store.path)
        return store

    def expand(self, pcoll: beam.PCollection) -> beam.PCollection:
        return pcoll | beam.Map(self._register_dataset_to_catalog)


@dataclass
class Copy(beam.PTransform):
    """Copy a store to a new location. If the target input is False, do nothing."""

    target: str

    def _copy(self, store: zarr.storage.FSStore) -> zarr.storage.FSStore:
        import os

        import gcsfs
        import zarr

        # We do need the gs:// prefix?
        # TODO: Determine this dynamically from zarr.storage.FSStore
        source = f'gs://{os.path.normpath(store.path)}/'  # FIXME more elegant. `.copytree` needs trailing slash
        if self.target is False:
            # dont do anything
            return store
        else:
            fs = gcsfs.GCSFileSystem()  # FIXME: How can we generalize this?
            fs.cp(source, self.target, recursive=True)
            # return a new store with the new path that behaves exactly like the input
            # to this stage (so we can slot this stage right before testing/logging stages)
            return zarr.storage.FSStore(self.target)

    def expand(self, pcoll: beam.PCollection) -> beam.PCollection:
        return pcoll | 'Copying Store' >> beam.Map(self._copy)


@dataclass
class InjectAttrs(beam.PTransform):
    inject_attrs: dict = None
    add_provenance: bool = True

    # add a post_init method to add the provenance attributes
    def __post_init__(self):
        if self.inject_attrs is None:
            self.inject_attrs = {}

        if self.add_provenance:
            git_url_hash = get_github_commit_url()
            gh_actions_url = get_github_actions_url()
            timestamp = datetime.now(timezone.utc).isoformat()
            provenance_dict = {
                'pangeo_forge_build_git_hash': git_url_hash,
                'pangeo_forge_gh_actions_url': gh_actions_url,
                'pangeo_forge_build_timestamp': timestamp,
            }
            self.inject_attrs.update(provenance_dict)

    def _update_zarr_attrs(self, store: zarr.storage.FSStore) -> zarr.storage.FSStore:
        # TODO: Can we get a warning here if the store does not exist?
        attrs = zarr.open(store, mode='a').attrs
        attrs.update(self.inject_attrs)
        # ? Should we consolidate here? We are explicitly doing that later...
        return store

    def expand(
        self, pcoll: beam.PCollection[zarr.storage.FSStore]
    ) -> beam.PCollection[zarr.storage.FSStore]:
        return pcoll | 'Injecting Attributes' >> beam.Map(self._update_zarr_attrs)



"""
utils that are specific to CMIP data management
"""

import logging
import warnings
from dataclasses import dataclass

import apache_beam as beam
import xarray as xr
import zarr
from dask.utils import parse_bytes
from dynamic_chunks.algorithms import (
    NoMatchingChunks,
    even_divisor_algo,
    iterative_ratio_increase_algo,
)
from pangeo_forge_recipes.transforms import Indexed, T

from leap_data_management_utils.bq_interfaces import CMIPBQInterface, IIDEntry
from leap_data_management_utils.cmip_testing import test_all

# TODO: I am not sure the chunking function belongs here, but it clutters the recipe and I did not want
# To open a whole file for this.
logger = logging.getLogger(__name__)


## Dynamic Chunking Wrapper
def dynamic_chunking_func(ds: xr.Dataset) -> dict[str, int]:
    logger.info(f'Input Dataset for dynamic chunking {ds =}')

    target_chunk_size = '150MB'
    target_chunks_aspect_ratio = {
        'time': 10,
        'x': 1,
        'i': 1,
        'ni': 1,
        'xh': 1,
        'nlon': 1,
        'lon': 1,  # TODO: Maybe import all the known spatial dimensions from xmip?
        'y': 1,
        'j': 1,
        'nj': 1,
        'yh': 1,
        'nlat': 1,
        'lat': 1,
    }
    size_tolerance = 0.5

    # Some datasets are smaller than the target chunk size and should not be chunked at all
    if ds.nbytes < parse_bytes(target_chunk_size):
        target_chunks = dict(ds.dims)

    else:
        try:
            target_chunks = even_divisor_algo(
                ds,
                target_chunk_size,
                target_chunks_aspect_ratio,
                size_tolerance,
                allow_extra_dims=True,
            )

        except NoMatchingChunks:
            warnings.warn(
                'Primary algorithm using even divisors along each dimension failed '
                'with. Trying secondary algorithm.'
                f'Input {ds=}'
            )
            try:
                target_chunks = iterative_ratio_increase_algo(
                    ds,
                    target_chunk_size,
                    target_chunks_aspect_ratio,
                    size_tolerance,
                    allow_extra_dims=True,
                )
            except NoMatchingChunks:
                raise ValueError(
                    'Could not find any chunk combinations satisfying '
                    'the size constraint with either algorithm.'
                    f'Input {ds=}'
                )
            # If something fails
            except Exception as e:
                raise e
        except Exception as e:
            raise e
    logger.info(f'Dynamic Chunking determined {target_chunks =}')
    return target_chunks


# ----------------------------------------------------------------------------------------------
# apache Beam stages
# ----------------------------------------------------------------------------------------------


@dataclass
class LogCMIPToBigQuery(beam.PTransform):
    """
    Logging stage for data written to zarr store
    """

    iid: str
    table_id: str
    retracted: bool = False
    tests_passed: bool = False

    def _log_to_bigquery(self, store: zarr.storage.FSStore) -> zarr.storage.FSStore:
        bq_interface = CMIPBQInterface(table_id=self.table_id)
        iid_entry = IIDEntry(
            iid=self.iid,
            store='gs://' + store.path,  # TODO: Get the full store path from the store object
            retracted=self.retracted,
            tests_passed=self.tests_passed,
        )
        bq_interface.insert_iid(iid_entry)
        return store

    def expand(self, pcoll: beam.PCollection) -> beam.PCollection:
        return pcoll | beam.Map(self._log_to_bigquery)


@dataclass
class Preprocessor(beam.PTransform):
    """
    Preprocessor for xarray datasets.
    Set all data_variables except for `variable_id` attrs to coord
    Add additional information

    """

    @staticmethod
    def _keep_only_variable_id(item: Indexed[T]) -> Indexed[T]:
        """
        Many netcdfs contain variables other than the one specified in the `variable_id` facet.
        Set them all to coords
        """
        index, ds = item
        print(f'Preprocessing before {ds =}')
        new_coords_vars = [var for var in ds.data_vars if var != ds.attrs['variable_id']]
        ds = ds.set_coords(new_coords_vars)
        print(f'Preprocessing after {ds =}')
        return index, ds

    @staticmethod
    def _sanitize_attrs(item: Indexed[T]) -> Indexed[T]:
        """Removes non-ascii characters from attributes see https://github.com/pangeo-forge/pangeo-forge-recipes/issues/586"""
        index, ds = item
        for att, att_value in ds.attrs.items():
            if isinstance(att_value, str):
                new_value = att_value.encode('utf-8', 'ignore').decode()
                if new_value != att_value:
                    print(
                        f'Sanitized datasets attributes field {att}: \n {att_value} \n ----> \n {new_value}'
                    )
                    ds.attrs[att] = new_value
        return index, ds

    def expand(self, pcoll: beam.PCollection) -> beam.PCollection:
        return (
            pcoll
            | 'Fix coordinates' >> beam.Map(self._keep_only_variable_id)
            | 'Sanitize Attrs' >> beam.Map(self._sanitize_attrs)
        )


@dataclass
class TestDataset(beam.PTransform):
    """
    Test stage for data written to zarr store
    """

    iid: str

    def _test(self, store: zarr.storage.FSStore) -> zarr.storage.FSStore:
        test_all(store, self.iid)
        return store

    def expand(self, pcoll: beam.PCollection) -> beam.PCollection:
        return pcoll | 'Testing - Running all tests' >> beam.Map(self._test)



import numpy as np
import pytest
import xarray as xr

from leap_data_management_utils.cmip_transforms import IIDEntry, dynamic_chunking_func


class TestIIDEntry:
    def test_IIDEntry(self):
        iid = 'CMIP6.CMIP.NCAR.CESM2.historical.r10i1p1f1.Amon.tas.gr1.v20190308'
        store = 'gs://cmip6/CMIP6/CMIP/NCAR/CESM2/historical/r10i1p1f1/Amon/tas/gr1/v20190308'
        retracted = False
        tests_passed = True
        entry = IIDEntry(iid, store, retracted, tests_passed)
        assert entry.iid == iid
        assert entry.store == store
        assert entry.retracted == retracted
        assert entry.tests_passed == tests_passed

    def test_too_long(self):
        iid = 'CMIP6.CMIP.NCAR.CESM2.historical.r10i1p1f1.Amon.tas.gr1.v20190308.extra'
        store = 'gs://cmip6/CMIP6/CMIP/NCAR/CESM2/historical/r10i1p1f1/Amon/tas/gr1/v20190308'
        retracted = False
        tests_passed = True
        with pytest.raises(ValueError):
            IIDEntry(iid, store, retracted, tests_passed)


class TestDynamicChunks:
    def test_too_small(self):
        ds = xr.DataArray(np.random.rand(4, 6)).to_dataset(name='data')
        chunks = dynamic_chunking_func(ds)
        assert chunks == {'dim_0': 4, 'dim_1': 6}


# TODO Its super hard to test anything involving big query, because AFAIK there is no way to mock it.



import pandas as pd

from leap_data_management_utils.cmip_catalog import bq_df_to_intake_esm


def test_bq_df_to_intake_esm():
    bq_df = pd.DataFrame(
        {
            'instance_id': [
                'CMIP6.AerChemMIP.MIROC.MIROC6.piClim-NTCF.sub-r1i1p1f1.Amon.tasmin.gn.v20190807',  # modified to test the sub_experiment split
                'CMIP6.AerChemMIP.MIROC.MIROC6.piClim-OC.r1i1p1f1.Amon.rlut.gn.v20190807',
            ],
            'store': [
                'gs://cmip6/CMIP6/AerChemMIP/MIROC/MIROC6/piClim-NTCF/r1i1p1f1/Amon/tasmin/gn/v20190807/',
                'gs://cmip6/CMIP6/AerChemMIP/MIROC/MIROC6/piClim-OC/r1i1p1f1/Amon/rlut/gn/v20190807/',
            ],
            'retracted': [False, False],
            'tests_passed': [True, True],
        }
    )
    intake_df = bq_df_to_intake_esm(bq_df)
    for c in intake_df.columns:
        print(c)
        print(intake_df[c].to_list())

    expected_intake_df = pd.DataFrame(
        {
            'activity_id': ['AerChemMIP', 'AerChemMIP'],
            'institution_id': ['MIROC', 'MIROC'],
            'source_id': ['MIROC6', 'MIROC6'],
            'experiment_id': ['piClim-NTCF', 'piClim-OC'],
            'member_id': ['sub-r1i1p1f1', 'r1i1p1f1'],
            'table_id': ['Amon', 'Amon'],
            'variable_id': ['tasmin', 'rlut'],
            'grid_label': ['gn', 'gn'],
            'sub_experiment_id': ['sub', 'none'],
            'variant_label': ['r1i1p1f1', 'r1i1p1f1'],
            'version': ['v20190807', 'v20190807'],
            'zstore': [
                'gs://cmip6/CMIP6/AerChemMIP/MIROC/MIROC6/piClim-NTCF/r1i1p1f1/Amon/tasmin/gn/v20190807/',
                'gs://cmip6/CMIP6/AerChemMIP/MIROC/MIROC6/piClim-OC/r1i1p1f1/Amon/rlut/gn/v20190807/',
            ],
        }
    )
    assert intake_df.equals(expected_intake_df)



import pandas as pd
from pangeo_forge_esgf.utils import CMIP6_naming_schema


def _maybe_prepend_dummy_dcpp(s: str):
    if '-' not in s:
        return 'none-' + s
    else:
        return s


def _maybe_join(iterable):
    assert len(iterable) == 2
    sub_experiment_id = iterable.iloc[0]
    variant_label = iterable.iloc[1]
    if sub_experiment_id != 'none':
        return f'{sub_experiment_id}-{variant_label}'
    else:
        return variant_label


def bq_df_to_intake_esm(df: pd.DataFrame) -> pd.DataFrame:
    df = df.copy()
    iid_facets = CMIP6_naming_schema.split('.')
    # some legit pandas magic here: https://stackoverflow.com/a/39358924
    df_out = df['instance_id'].str.split('.', expand=True)
    df_out = df_out.rename(columns={i: f for i, f in enumerate(iid_facets)})
    df_out['zstore'] = df['store']
    df_out[['sub_experiment_id', 'variant_label']] = (
        df_out['member_id']
        .apply(lambda s: _maybe_prepend_dummy_dcpp(s))
        .str.split('-', expand=True)
    )
    # df.replace('dummy', np.nan, inplace=True)
    # order the columns in the order of the old csv (not sure if this is necessary)
    df_out = df_out[
        [
            'activity_id',
            'institution_id',
            'source_id',
            'experiment_id',
            'member_id',
            'table_id',
            'variable_id',
            'grid_label',
            'sub_experiment_id',
            'variant_label',
            'version',
            'zstore',
        ]
    ]
    return df_out


def intake_esm_df_to_bq_df(df: pd.DataFrame) -> pd.DataFrame:
    # now remove the ones already in the pangeo catalog
    df = df.copy()
    df['member_id'] = df[['sub_experiment_id', 'variant_label']].agg(_maybe_join, axis=1)
    df['version'] = 'v' + df['version'].astype(str)
    df['instance_id'] = (
        df[
            [
                'activity_id',
                'institution_id',
                'source_id',
                'experiment_id',
                'member_id',
                'table_id',
                'variable_id',
                'grid_label',
                'version',
            ]
        ]
        .astype(str)
        .agg('.'.join, axis=1)
        .tolist()
    )
    df['instance_id'] = 'CMIP6.' + df['instance_id']
    df['store'] = df['zstore']
    # add current time as bigquery timestamp
    df['timestamp'] = pd.Timestamp.now(tz='UTC')
    df = df[['instance_id', 'store', 'timestamp']]
    return df



from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from google.api_core.exceptions import NotFound
from google.cloud import bigquery
from pangeo_forge_esgf.utils import CMIP6_naming_schema
from tqdm.auto import tqdm


@dataclass
class BQInterface:
    """Class to read/write information from BigQuery table
    :param table_id: BigQuery table ID
    :param client: BigQuery client object
    :param result_limit: Maximum number of results to return from query
    """

    table_id: str
    client: Optional[bigquery.client.Client] = None
    result_limit: Optional[int] = 10
    schema: Optional[list] = None

    def __post_init__(self):
        # TODO how do I handle the schema? This class could be used for any table, but for
        # TODO this specific case I want to prescribe the schema
        # for now just hardcode it
        if not self.schema:
            self.schema = [
                bigquery.SchemaField('dataset_id', 'STRING', mode='REQUIRED'),
                bigquery.SchemaField('dataset_url', 'STRING', mode='REQUIRED'),
                bigquery.SchemaField('timestamp', 'TIMESTAMP', mode='REQUIRED'),
            ]
        if self.client is None:
            self.client = bigquery.Client()

        # check if table exists, otherwise create it
        try:
            self._get_table()
        except NotFound:
            self.create_table()

    def create_table(self) -> bigquery.table.Table:
        """Create the table if it does not exist"""
        print(f'Creating {self.table_id =}')
        table = bigquery.Table(self.table_id, schema=self.schema)
        self.client.create_table(table)  # Make an API request.

    def _get_table(self) -> bigquery.table.Table:
        """Get the table object"""
        return self.client.get_table(self.table_id)

    def insert(self, fields: dict = {}):
        timestamp = datetime.now().isoformat()

        rows_to_insert = [
            fields | {'timestamp': timestamp}  # timestamp is always overridden
        ]

        errors = self.client.insert_rows_json(self._get_table(), rows_to_insert)
        if errors:
            raise RuntimeError(f'Error inserting row: {errors}')

    def catalog_insert(self, dataset_id: str, dataset_url: str, extra_fields: dict = {}):
        rows_to_insert = [
            {
                'dataset_id': dataset_id,
                'dataset_url': dataset_url,
            }
            | extra_fields
        ]
        self.insert(rows_to_insert)

    def _get_query_job(self, query: str) -> bigquery.job.query.QueryJob:
        return self.client.query(query)

    def get_all(self) -> list[bigquery.table.Row]:
        """Get all rows in the table"""
        query = f"""
        SELECT * FROM {self.table_id};
        """
        results = self._get_query_job(query)
        return results.to_dataframe()

    def get_latest(self) -> list[bigquery.table.Row]:
        """Get the latest row for all iids in the table"""
        # adopted from https://stackoverflow.com/a/1313293
        query = f"""
        WITH ranked_iids AS (
        SELECT i.*, ROW_NUMBER() OVER (PARTITION BY instance_id ORDER BY timestamp DESC) AS rn
        FROM {self.table_id} AS i
        )
        SELECT * FROM ranked_iids WHERE rn = 1;
        """
        results = self._get_query_job(query)
        return results.to_dataframe().drop(columns=['rn'])


@dataclass
class IIDEntry:
    """Single row/entry for an iid
    :param iid: CMIP6 instance id
    :param store: URL to zarr store
    """

    iid: str
    store: str  # TODO: should this allow other objects?
    retracted: bool
    tests_passed: bool

    # Check if the iid conforms to a schema
    def __post_init__(self):
        schema = CMIP6_naming_schema
        facets = self.iid.split('.')
        if len(facets) != len(schema.split('.')):
            raise ValueError(f'IID does not conform to CMIP6 {schema =}. Got {self.iid =}')
        assert self.store.startswith('gs://')
        assert self.retracted in [True, False]
        assert self.tests_passed in [True, False]

        # TODO: Check each facet with the controlled CMIP vocabulary

        # TODO Check store validity?


@dataclass
class IIDResult:
    """Class to handle the results pertaining to a single IID."""

    results: bigquery.table.RowIterator
    iid: str

    def __post_init__(self):
        if self.results.total_rows > 0:
            self.exists = True
            self.rows = [r for r in self.results]
            self.latest_row = self.rows[0]
        else:
            self.exists = False


class CMIPBQInterface(BQInterface):
    """Class to read/write information from BigQuery table
    :param table_id: BigQuery table ID
    :param client: BigQuery client object
    :param result_limit: Maximum number of results to return from query
    """

    def __post_init__(self):
        # TODO how do I handle the schema? This class could be used for any table, but for
        # TODO this specific case I want to prescribe the schema
        # for now just hardcode it
        if not self.schema:
            self.schema = [
                bigquery.SchemaField('instance_id', 'STRING', mode='REQUIRED'),
                bigquery.SchemaField('store', 'STRING', mode='REQUIRED'),
                bigquery.SchemaField('timestamp', 'TIMESTAMP', mode='REQUIRED'),
                bigquery.SchemaField('retracted', 'BOOL', mode='REQUIRED'),
                bigquery.SchemaField('tests_passed', 'BOOL', mode='REQUIRED'),
            ]
        super().__post_init__()

    def _get_timestamp(self) -> str:
        """Get the current timestamp"""
        return datetime.datetime.utcnow().isoformat()

    def insert_iid(self, IID_entry):
        """Insert a row into the table for a given IID_entry object"""
        fields = {
            'instance_id': IID_entry.iid,
            'store': IID_entry.store,
            'retracted': IID_entry.retracted,
            'tests_passed': IID_entry.tests_passed,
            'timestamp': self._get_timestamp(),
        }
        self.insert(fields)

    def insert_multiple_iids(self, IID_entries: list[IIDEntry]):
        """Insert multiple rows into the table for a given list of IID_entry objects"""
        # FIXME This repeats a bunch of code from the parent class .insert() method
        timestamp = self._get_timestamp()
        rows_to_insert = [
            {
                'instance_id': IID_entry.iid,
                'store': IID_entry.store,
                'retracted': IID_entry.retracted,
                'tests_passed': IID_entry.tests_passed,
                'timestamp': timestamp,
            }
            for IID_entry in IID_entries
        ]
        errors = self.client.insert_rows_json(self._get_table(), rows_to_insert)
        if errors:
            raise RuntimeError(f'Error inserting row: {errors}')

    def _get_iid_results(self, iid: str) -> IIDResult:
        # keep this in case I ever need the row index again...
        # query = f"""
        # WITH table_with_index AS (SELECT *, ROW_NUMBER() OVER ()-1 as row_index FROM `{self.table_id}`)
        # SELECT *
        # FROM `table_with_index`
        # WHERE instance_id='{iid}'
        # """
        """Get the full result object for a given iid"""
        query = f"""
        SELECT *
        FROM `{self.table_id}`
        WHERE instance_id='{iid}'
        ORDER BY timestamp DESC
        LIMIT {self.result_limit}
        """
        results = self._get_query_job(
            query
        ).result()  # TODO: `.result()` is waiting for the query. Should I do this here?
        return IIDResult(results, iid)

    def iid_exists(self, iid: str) -> bool:
        """Check if iid exists in the table"""
        return self._get_iid_results(iid).exists

    def _iid_list_exists_batch(self, iids: list[str]) -> list[str]:
        """More efficient way to check if a list of iids exists in the table
        Passes the entire list to a single SQL query.
        Returns a list of iids that exist in the table
        ```
        """
        if len(iids) > 10000:
            raise ValueError('List of iids is too long. Please work in batches.')

        # source: https://stackoverflow.com/questions/26441928/how-do-i-check-if-multiple-values-exists-in-database
        query = f"""
        SELECT instance_id, store
        FROM {self.table_id}
        WHERE instance_id IN ({",".join([f"'{iid}'" for iid in iids])})
        """
        results = self._get_query_job(query).result()
        # this is a full row iterator, for now just return the iids
        return list(set([r['instance_id'] for r in results]))

    def iid_list_exists(self, iids: list[str]) -> list[str]:
        """More efficient way to check if a list of iids exists in the table
        Passes the entire list in batches into SQL querys for maximum efficiency.
        Returns a list of iids that exist in the table
        """

        # make batches of the input, since bq cannot handle more than 10k elements here
        iids_in_bq = []
        batchsize = 10000
        iid_batches = [iids[i : i + batchsize] for i in range(0, len(iids), batchsize)]
        for iids_batch in tqdm(iid_batches):
            iids_in_bq_batch = self._iid_list_exists_batch(iids_batch)
            iids_in_bq.extend(iids_in_bq_batch)
        return iids_in_bq



"""
This module contains functions for testing CMIP6 zarr stores.
Its a bit awkward to put them here, but I want to use these functions both in
the apache-beam pipelines and in scripts that run occasionally to re-check all
stores. Due to the way that pangeo-forge works right now, we cannot separate
these in modules in https://github.com/leap-stc/cmip6-leap-feedstock
"""

# ---------------------------------------------------------------------------------------------------------------------------------------------------------
# Test functions (I will need to use these outside too)
# --------------------------------------------------------------------------------------------------------------------------------------------------------
import xarray as xr
import zarr
from pangeo_forge_esgf.utils import CMIP6_naming_schema


def open_dataset(store: zarr.storage.FSStore) -> xr.Dataset:
    return


def test_open_store(store: zarr.storage.FSStore, verbose):
    if verbose:
        print('Testing - Open Store')
        print(f'Written path: {store.path =}')
    ds = xr.open_dataset(
        store.fs.get_mapper(store.path),
        engine='zarr',
        chunks={},
        use_cftime=True,
        consolidated=True,
    )
    if verbose:
        print(ds)
    return ds


def test_time(ds: xr.Dataset, verbose):
    """
    Check time dimension
    For now checks:
    - That time increases strictly monotonically
    - That no large gaps in time (e.g. missing file) are present
    """
    if verbose:
        print('Testing - Time Dimension')
    time_diff = ds.time.diff('time').astype(int)
    # assert that time increases monotonically
    if verbose:
        print(time_diff)
    assert (time_diff > 0).all()

    # assert that there are no large time gaps
    mean_time_diff = time_diff.mean()
    normalized_time_diff = abs((time_diff - mean_time_diff) / mean_time_diff)
    assert (normalized_time_diff < 0.05).all()


def test_attributes(ds: xr.Dataset, iid: str, verbose):
    if verbose:
        print('Testing - Attributes')

    # check completeness of attributes
    iid_schema = CMIP6_naming_schema
    # The cmip datasets do not all have 'member_id' as an attribute.
    # Member id is composed of two (I think) required attributes as:
    # <sub_experiment_id>-<variant_label>
    # See https://docs.google.com/document/d/1h0r8RZr_f3-8egBMMh7aqLwy3snpD6_MrDz1q8n5XUk/edit for more info
    iid_schema = iid_schema.replace('.member_id.', '.variant_label.sub_experiment_id.')
    for facet_value, facet in zip(iid.split('.'), iid_schema.split('.')):
        if 'version' not in facet:  # (TODO: Why is the version not in all datasets?)
            if verbose:
                print(f'Checking {facet = } in dataset attributes')
            assert ds.attrs[facet] == facet_value


def test_all(store: zarr.storage.FSStore, iid: str, verbose=True) -> zarr.storage.FSStore:
    ds = test_open_store(store, verbose=verbose)
    test_time(ds, verbose=verbose)
    test_attributes(ds, iid, verbose=verbose)
    return store
